classdef Kernel < handle
    % probabilistic kernel
    %
    % author: Aurlien Garivier
    % reference: [1] Perfect Simulation Scheme For Processes With Long Memory :
    %    A ``Coupling Into And From The Past'' Algorithm, by A. Garivier   
    
    properties
        K % alphabet size
    end
    
    methods
        function p = subDist(obj, s)
            % given a context s, p(x) = min(P(x | z) : z in T(s))
        end
        
        function f = transition(obj, u)
            % VIRTUAL FUNCTION: to be defined in subclasses
            % given u in [0,1[, returns the PCfunction of the context 
            % giving the next symbol            
        end
        
        function [X, t] = CIAFTP(obj, n, verbose)
            % sample X = [X(n),...,X(-1)] from stationary distribution
            % using the CIAFTP simulation scheme of [1]
            % if verbose is set to true, the algorithm shows step by step
            % the evolution of function G
            if nargin<3, verbose = false; end
            if nargin<2, n = -1; end
            if n>0, n=-n; end % unnecessarily prudent?

            H = Context.PCfunction.identity(obj.K); t = -1;
            while ~H.isConstant()
                oldH = H.copy(); % can be removed: useful only for verbose mode
                t = t - 1;
                U = rand; 
                f = obj.transition(U);
                leaves = f.CSD.getLeaves();
                m = length(leaves);
                nH = Context.PCfunction(f.CSD.copy(), cell(1,length(f.CSD.father)));
                for k=1:m
                    w = [f.CSD.getWord(leaves(k)), f.values{leaves(k)}];
                    [s, isLeaf] = H.CSD.getSuffix(w);
                    if isLeaf
                        nH.values{leaves(k)} = H.values{s};
                    else
                        [subT, nodes] = H.CSD.getSubTrie(s);
                        val = H.getValues(nodes);
                        nH.insertDetails(leaves(k), subT, val);
                    end
                end
                if verbose
                    disp(['t = ', num2str(t), ', U = ', num2str(U)]);    
                    disp(['before pruning, H has depth ', num2str(nH.CSD.getDepth())]);
                end
                nH.prune();
                if t>=n
                    if nH.isConstant()
                        H = Context.PCfunction.identity(obj.CSD.K);
                        temp = cellfun(@(a) [a, nH.values{1}], {H.values{2:end}}, 'UniformOutput', false);
                        [H.values{2:end}] = temp{:};
                    else
                        H = nH; 
                        for x = 1:H.CSD.K
                            leaves = H.CSD.getLeaves(H.CSD.getChild(H.CSD.getRoot(), x));
                            temp = cellfun(@(a) [x, a], {H.values{leaves}}, 'UniformOutput', false);
                            [H.values{leaves}] = temp{:};
                        end
                    end
                else
                    H = nH.copy();
                    %H = nH; % Warning: not copying, but this is ok here (nH is not used later)
                end
                
                if verbose                 
                    subplot(2,2,1);
                    f.plot();% pause
                    subplot(2,2,3); oldH.plot();                
                    subplot(2,2,2); 
                    nH.plot(); %pause;
                    disp(['after pruning, H has depth ', num2str(nH.CSD.getDepth())]);
                    subplot(2,2,4);
                    H.plot(); 
                    pause;
                end
            end
            X = H.values{1};
        end
        
    end
    
end

